//==============================================================================
//== General Console Application utilities
//== Eric Boisvert
//== 25 mai 2007
//==
//== Initialisation automatique des handles
//== InputHan et OutputHan vers StdIn et SdtOut
//==
//== En vedette dans cette unit:
//== CLS() <- c'est classique...
//== SetFullScreen() <- Force le mode plein cran de l'application console
//== Sauvegarde et rappele de tout l'cran sur une pile crer automatiquement
//==    -PushScreen et PopScreen
//== Controle du curseur
//==    -ShowCursor et HideCursor
//== Fontions pour affichage de fentres genre Poppup-menu
//==    -PushWindow et PopWindow
//== Effets Spciaux
//==    -MakeFlash() (une zone ou un seul caratre)
//==============================================================================
Unit ConsoleScreen;

Interface

Uses sysutils, windows, contnrs,classes;


{(*} //START to Prevent Delphi Formater change here
{==============================================================================}
{                                }Const {                                      }
{==============================================================================}
  CONSOLE_FULLSCREEN_MODE = 1; // Mode plein cran
  CONSOLE_WINDOWED_MODE   = 2; // Mode console ('fentr').

  SCREENWITH    = 80;
  SCREENHEIGHT  = 50;
  //======================================//
  //== Vieux caractres Graphiques DOS ===//
  NECORNER    = #187; //
  SECORNER    = #188; //
  SOCORNER    = #200; //
  NOCORNER    = #201; //
  UPDOWN      = #186; //
  LEFTRIGHT   = #205; //
  TUP         = #202; //
  TDOWN       = #203; //
  TLEFT       = #185; //
  TRIGHT      = #204; //
  TITLE_LEFT  = #181; //
  TITLE_RIGHT = #198; //

  SPACE       = #032; //

  //== Couleur blanche pour les menu ==
  MENU_NORMAL = FOREGROUND_RED + FOREGROUND_GREEN + FOREGROUND_BLUE;
  MENU_NORMALHIGHLIGHT = FOREGROUND_RED +
                         FOREGROUND_GREEN +
                         FOREGROUND_BLUE + FOREGROUND_INTENSITY;

  //== Flashing Effect ==
  MaxFlashSeq = 9;
  FlashSeq: Array[0..MaxFlashSeq] Of Word =
   (FOREGROUND_RED + FOREGROUND_GREEN + FOREGROUND_BLUE + FOREGROUND_INTENSITY,
    FOREGROUND_RED + FOREGROUND_BLUE,
    FOREGROUND_RED + FOREGROUND_BLUE + FOREGROUND_INTENSITY,
    FOREGROUND_RED + FOREGROUND_GREEN,
    FOREGROUND_RED + FOREGROUND_GREEN + FOREGROUND_INTENSITY,
    FOREGROUND_GREEN + FOREGROUND_BLUE,
    FOREGROUND_GREEN + FOREGROUND_BLUE + FOREGROUND_INTENSITY,
    FOREGROUND_RED,
    FOREGROUND_RED + FOREGROUND_INTENSITY,
    FOREGROUND_BLUE);

 {*)}//STOP to Prevent Delphi Formater change here

{==============================================================================}
{==                            ==}Var {==                                    ==}
{==============================================================================}

  //== Stdin et StdOut handle initialization section ==
  OutputHan: THandle;
  InputHan: THandle;
  //== Pile pour la sauvegarde complte de l'cran ==
  ScreenStack: TStack;
  //== Index pour la couleur courrante pour le flashing effect ==
  FlashIndex: integer = 0;


{==============================================================================}
{                                }Type {                                       }
{==============================================================================}
  //== Screen Buffer ==
  PScrBuf = ^TScrBuf;
  TScrBuf = Array[0..SCREENWITH * SCREENHEIGHT] Of TCharInfo;



{========================== Public Functions ==================================}
//== Make console application in Full Screen ==
Function SetFullScreen: Boolean;
//== Clear screen ==
Procedure CLS();
//== Hide-Show cursor ==
Procedure HideCursor();
Procedure ShowCursor(Pos: TCOORD);

//== Save All screen on stack ==
Procedure PushScreen();
//== Restore a screen from stack ==
Procedure PopScreen();
//== Create a popup menu window ==
Function PushWindow(Width, Height: integer; Title: String; Attributes: Word): TCOORD;
//== Delete a popup menu window ==
Procedure PopWindow();
//== Flashing effect Single char or zone ==
Procedure MakeFlash(Pos: TCOORD); overload;
Procedure MakeFlash(LeftTopPos: TCOORD; RightBottomPos: TCOORD); overload;

Procedure HideMouseCursor();


//== Debug only function that display all caracters on screen ==
Procedure FillCharDisplay(FromChar: byte; ToChar: byte);





{==============================================================================}
{                            }Implementation {                                 }
{==============================================================================}



{========================= Fonction Extrieur =================================}
//== Pour Mettre en mode plein cran.
Function SetConsoleDisplayMode(ConsoleOutput: THandle; Flags: DWORD;
  NewScreenBufferDimensions: PCoord): BOOL;
  stdcall; external 'kernel32.dll';



{==============================================================================}
{=== Met l'application console en mode Pleine cran ou en mode fentr      ===}
{=== Source: www.developpez.com FAQ de delphi                               ===}
Function SetConsoleScreen(Mode: Byte): Boolean;
Var
  ConsoleHandle: THandle;
Begin
  Result := False;
  ConsoleHandle := GetStdHandle(STD_OUTPUT_HANDLE);
  If ConsoleHandle <> INVALID_HANDLE_VALUE Then
  Begin
    Result := SetConsoleDisplayMode(ConsoleHandle, Mode, Nil) = False
  End;
End;


{==============================================================================}
{=== Met l'application console en mode Pleine cran ==}
Function SetFullScreen: Boolean;
Begin
  Result := SetConsoleScreen(CONSOLE_FULLSCREEN_MODE);
End;


{==============================================================================}
{=== Efface tout l'cran                                                    ===}
{=== Source: www.developpez.com FAQ de delphi                               ===}
Procedure CLS();
Var
  Pos: TCoord;
  dw: DWord;
Begin
  Pos.X := 0;
  Pos.Y := 0;
  FillConsoleOutputCharacter(OutputHan,SPACE,SCREENWITH * SCREENHEIGHT,Pos,dw);
  FillConsoleOutputAttribute(OutputHan, 0, SCREENWITH * SCREENHEIGHT, Pos, dw);
End;


{==============================================================================}
{=== Cache le curseur                                                       ===}
{=== Source: www.developpez.com FAQ de delphi                               ===}
Procedure HideCursor();
Var
  ci: TConsoleCursorInfo;
Begin
  GetConsoleCursorInfo(OutputHan, ci);
  ci.bVisible := FALSE;
  SetConsoleCursorInfo(OutputHan, ci);
End;


{==============================================================================}
{=== Affiche le curseur  la position dsire                               ===}
Procedure ShowCursor(Pos: TCOORD);
Var
  ci: TConsoleCursorInfo;
Begin
  SetConsoleCursorPosition(OutputHan, Pos);
  GetConsoleCursorInfo(OutputHan, ci);
  ci.bVisible := TRUE;
  SetConsoleCursorInfo(OutputHan, ci);
End;


{==============================================================================}
{=== Sauvegarde entirement l'cran dans sur une pile                       ===}
Procedure PushScreen();
Var
  dwBufferSize: COORD;
  dwBufferCoord: COORD;
  smallRect: TSmallRect;
  pScreen: PScrBuf;
Begin
  //== limite du buffer ==
  dwBufferSize.X := SCREENWITH;
  dwBufferSize.Y := SCREENHEIGHT;
  //== coordonnes de destination dans le buffer ==
  dwBufferCoord.X := 0;
  dwBufferCoord.Y := 0;

  //== Coordonne de lecture (tout l'cran) ==
  SmallRect.Top := 0;
  SmallRect.Left := 0;
  SmallRect.Right := SCREENWITH;
  SmallRect.Bottom := SCREENHEIGHT;
  //== Cration dynamique du buffer ==
  New(pScreen);
  Try
    ReadConsoleOutput(OutputHan,
      pScreen,
      dwBufferSize,
      dwBufferCoord,
      smallRect);
  Finally
    //== Sauvegarde de l'cran sur la pile ==
    ScreenStack.Push(pScreen);
  End;
End;


{==============================================================================}
{=== Rcupre un cran sur la pile et la re-dessine  l'cran               ===}
Procedure PopScreen();
Var
  dwBufferSize: COORD;
  dwBufferCoord: COORD;
  smallRect: TSmallRect;
  pScreen: PScrBuf;
Begin
  //== verifie qu'il y a au moins 1 cran disponible dans la pile... ==
  If ScreenStack.Count = 0 Then exit;

  //== Rcupration du buffer cran sur la pile ==
  pScreen := ScreenStack.Pop;
  If pScreen <> Nil Then
  Begin
    //== Grosseur du buffer ==
    dwBufferSize.X := SCREENWITH;
    dwBufferSize.Y := SCREENHEIGHT;
    //== Commence la lecture  la coordonne (0,0) ==
    dwBufferCoord.X := 0;
    dwBufferCoord.Y := 0;
    //== Coordonnes de destination pour l'criture de l'cran ==
    SmallRect.Top := 0;
    SmallRect.Left := 0;
    SmallRect.Right := SCREENWITH;
    SmallRect.Bottom := SCREENHEIGHT;
    Try
      //== Transcrit le buffer dans l'cran ==
      WriteConsoleOutput(OutputHan,
        pScreen,
        dwBufferSize,
        dwBufferCoord,
        smallRect);
    Finally
      //== Libration de la mmoire allou dynamiquement par PushScreen() ==
      Dispose(pScreen);
    End;
  End;
End;


//==============================================================================
//== Sauvegarde l'ecran et Affiche un Popup Menu  l'cran.
//== Retourne les coordonnes de la 1ere ligne vide de la fentre.
//== Les menu sont toujours centrs  l'cran.
Function PushWindow(Width, Height: integer; Title: String; Attributes: Word): TCOORD;
Var
  Pos: TCoord;
  strLine: String;
  i: Integer;
  dw: DWord;
Begin
  PushScreen();
  SetConsoleTextAttribute(OutputHan, Attributes);
  //== Check for range ==
  If Width < 4 Then Width := 4;
  If Height < 1 Then Height := 1;
  If Width > SCREENWITH Then Width := SCREENWITH;
  If Height > SCREENHEIGHT - 2 Then Height := SCREENHEIGHT - 2;
  Pos.X := (SCREENWITH Div 2) - (Width Div 2);
  Pos.Y := (SCREENHEIGHT Div 2) - (Height Div 2);
  //== Title ==
  strLine := StringOfChar(LEFTRIGHT, Width);
  strLine[1] := NOCORNER;
  strLine[Width] := NECORNER;
  If Title <> '' Then
  Begin
    If Length(Title) > (Width - 4) Then
    Begin
      Title := Copy(Title, 1, Width - 4);
    End;
    Title := TITLE_LEFT + '' + Title + '' + TITLE_RIGHT;
    strmove(@strLine[((Width Div 2) - (Length(Title) Div 2)) + 1], PChar(Title), Length(Title));
  End;
  SetConsoleCursorPosition(OutputHan, Pos);
  WriteConsole(OutputHan, PChar(strLine), length(strLine), dw, Nil);
  result := Pos;
  inc(Result.X);
  inc(Result.Y);
  //== Mid window ==
  strLine := StringOfChar(SPACE, Width);
  strLine[1] := UPDOWN;
  strLine[Width] := UPDOWN;
  For i := 0 To height - 2 Do
  Begin
    inc(Pos.y);
    SetConsoleCursorPosition(OutputHan, Pos);
    WriteConsole(OutputHan, PChar(strLine), length(strLine), dw, Nil);
  End;
  //== Bottom ==
  strLine := StringOfChar(LEFTRIGHT, Width);
  strLine[1] := SOCORNER;
  strLine[Width] := SECORNER;
  inc(Pos.y);
  SetConsoleCursorPosition(OutputHan, Pos);
  WriteConsole(OutputHan, PChar(strLine), length(strLine), dw, Nil);
End;


//==============================================================================
//== Efface un PopupMenu
Procedure PopWindow();
Begin
  PopScreen();
End;


//==============================================================================
//== Fait changer de couleur un seul caractre  l'cran
Procedure MakeFlash(Pos: TCOORD);
Var
  dw: dword;
Begin
  If FlashIndex > MaxFlashSeq Then FlashIndex := 0;
  FillConsoleOutputAttribute(OutputHan, FlashSeq[FlashIndex], 1, Pos, dw);
  inc(FlashIndex);
End;


//==============================================================================
//== Fait changer de couleur une zone de caratre  l'cran.
//== La zone est dfini  l'aide de 2 TCOORD
//== LeftTopPos reprsente le coin Gauche en Haut de la zone.
//== RightBottomPos reprsente le coin Droit en Bas de la zone.
//== Le colorisation se fait ligne par ligne dans la zone.
Procedure MakeFlash(LeftTopPos: TCOORD; RightBottomPos: TCOORD); overload;
Var
  dw: dword;
  i: integer;
  Row: integer;
Begin
  //== Verifie la limite de l'index de l'effet de FLASH ==
  If FlashIndex > MaxFlashSeq Then FlashIndex := 0;

  //== Calcule le nombre de ligne  colorier ==
  Row := RightBottomPos.y - LeftTopPos.y;
  For i := 0 To Row Do
  Begin
    //== Colorisation d'une ligne ==
    FillConsoleOutputAttribute(OutputHan, FlashSeq[FlashIndex],
      RightBottomPos.x - LeftTopPos.x + 1,
      LeftTopPos, dw);
    //== Changement de ligne ==
    inc(LeftTopPos.Y);
  End;
  //== change la couleur pour le prochain call ==
  inc(FlashIndex);
End;


Procedure RestrictCursorToBottomRight(winRect:TRect);
Var
  CursorClipArea: TRect;
Begin
  CursorClipArea := Bounds(winRect.Right , winRect.BOTTOM, 1,1);
  Windows.ClipCursor(@CursorClipArea)
End;

Procedure HideMouseCursor();
var
  hwnd:THandle;
  WinRect:TRect;
begin
  hwnd:=GetDesktopWindow();
  GetClientRect(hwnd,WinRect);
  RestrictCursorToBottomRight(WinRect);
end;

//==============================================================================
//== DEBUG-TEST ONLY
//== Affiche tout les caratres  l'cran...
//== Pratique pour choisir des cractres...
Procedure FillCharDisplay(FromChar: byte; ToChar: byte);
Var
  dwBufferSize: COORD;
  dwBufferCoord: COORD;
  smallRect: TSmallRect;
  pScreen: PScrBuf;
  i: integer;
Begin
  //== Cration dynamique d'un buffer d'cran ==
  New(pScreen);
  Try
    //== Vide compltement le buffer ==
    For i := 0 To SCREENWITH * SCREENHEIGHT Do
    Begin
      pScreen^[i].AsciiChar := SPACE;
      pScreen^[i].Attributes := FOREGROUND_RED + FOREGROUND_GREEN + FOREGROUND_BLUE + FOREGROUND_INTENSITY;
    End;
    //== Affiche seulement la plage dsir ==
    For i := FromChar To ToChar Do
    Begin
      pScreen^[i].AsciiChar := chr(i);
    End;

    //== Limite du buffer ==
    dwBufferSize.X := SCREENWITH;
    dwBufferSize.Y := SCREENHEIGHT;
    //== Debut de lecture du buffer  la coordonne (0,0);
    dwBufferCoord.X := 0;
    dwBufferCoord.Y := 0;

    //== Destination  l'cran
    SmallRect.Top := 0;
    SmallRect.Left := 0;
    SmallRect.Right := SCREENWITH;
    SmallRect.Bottom := SCREENHEIGHT;
    //== Ecriture du buffer dans l'cran ==
    WriteConsoleOutput(OutputHan,
      pScreen,
      dwBufferSize,
      dwBufferCoord,
      smallRect);
  Finally
    //== Libration de la mmoire...==
    dispose(pScreen);
  End;
End;




//==============================================================================
//== Creation de la pile d'cran
//== Cette fonction est appel par la section Initialization
Procedure AutoInitScreenStack();
Begin
  ScreenStack := TStack.Create;
End;

//==============================================================================
//== Rcupration des handles StdIn et StdOut
//== Cette fonction est appel par la section Initialization
Procedure InitStdInOut();
Begin
  OutputHan := GetStdHandle(STD_OUTPUT_HANDLE);
  InputHan := GetStdHandle(STD_INPUT_HANDLE);
End;


//==============================================================================
//== Vide la pile d'cran et destruction de la pile.
//== Cette fonction est appel par la section Finalization
Procedure AutoDestroyScreenStack();
Var
  pScreen: PScrBuf;
Begin
  While ScreenStack.Count <> 0 Do
  Begin
    pScreen := ScreenStack.Pop;
    Dispose(pScreen);
  End;
  ScreenStack.Free;
End;


//==============================================================================
//=== section Initialization et Finalization de l'unit
//==============================================================================
Initialization
  AutoInitScreenStack();
  InitStdInOut();

Finalization
  AutoDestroyScreenStack();

End.

