Unit HighScore;

Interface

Uses Global, Menu, ConsoleScreen, FTPWininet,
  Sysutils, windows, classes, ActiveX, ComObj;

{==============================================================================}
{                                }Const {                                     }
{==============================================================================}
  HI_SC_NAME = 'HS_NET';


{========================== Public Functions ==================================}
function CheckHiScore(Player:integer;Score:integer):boolean;
Procedure GetInternetHS(LocalPath: String);





{==============================================================================}
{                            }Implementation {                                 }
{==============================================================================}



//==============================================================================
//== Creation d'un Indentifigateur Unique Global (GUID)
Function CreateGuid(): String;
Var
  ID: TGUID;
Begin
  Result := '';
  If CoCreateGuid(ID) = S_OK Then
    Result := GUIDToString(ID);
End;


//==============================================================================
//== Tentative pour effectu la lecture des High-score sur internet
//== Si ca fonctionne, on crase les High-Score locaux avec ceux obtenus sur
//== internet.
Procedure GetInternetHS(LocalPath: String);
Var
  Pos: TCoord;
  FileName:string;
Begin
  PushScreen();
  INTERNET_HS_ENABLE:=FALSE;
  Try
    Pos.X := 1;
    Pos.y := 1;
    SetConsoleCursorPosition(OutputHan, Pos);
    Writeln('Please wait...Try to get internet High Score...');

    If FTPFileExist(HI_SC_NAME) Then
    Begin
      FileName:=IncludeTrailingBackslash(LocalPath) + HI_SC_NAME;
      GetFTPFile(HI_SC_NAME,FileName);
      hs.LoadFromFile(FileName);
      DeleteFile(PChar(FileName));
      INTERNET_HS_ENABLE := TRUE;
    End;
  Finally
    PopScreen();
  End;
End;

//==============================================================================
//== Tentative pour effectu la sauvegarde des High-score sur internet
Procedure SaveInternetHS();
Var
  HS_TempFileName: String;
  SaveOk: boolean;
  i: integer;
  Pos: TCoord;
  GUIDFileName:string;
Begin
  //== Sauvegarde l'cran... ==
  PushScreen();
  Try
    Pos.X := 1;
    Pos.y := 2;
    SetConsoleCursorPosition(OutputHan, Pos);
    Writeln('Please wait...Try to save internet High Score...');
    //== ca nous prend un nom de fichier Unique avec CreateGuid() ==
    GUIDFileName:=CreateGuid();
    HS_TempFileName :=IncludeTrailingBackslash(ExtractFilePath(ParamStr(0))) +GUIDFileName ;
    hs.SaveToFile(HS_TempFileName);
    SaveOk := FALSE;
    //== Envoi sur internet ==
    If SendFTPFile(HS_TempFileName) = TRUE Then
    Begin
      //== Le serveur devrait traiter le fichier recu...et l'ffacer ==
      //== On attend donc qu'il ne soit plus sur le site FTP ==
      For i := 0 To 5 Do
      Begin
        sleep(500);
        If Not FTPFileExist(GUIDFileName) Then
        Begin
          //== Le fichier a disparue, donc ==
          //== l'operation de sauvegarde c'est bien pass... ==
          SaveOk := TRUE;
          break;
        End
      End;
    End;
    //== Nettoyage du disque ==
    if FileExists(HS_TempFileName) then DeleteFile(PChar(HS_TempFileName));
    If SaveOk Then
    Begin
      //=== Recupre le nouveau fichier de High-score sur le Net ===
      //=== Il y a peut-tre d'autre joueur qui on soumis des High-Score ==
      GetInternetHS(ExtractFilePath(ParamStr(0)))
    End;
  Finally
    //== Remet l'cran dans son tat prcdent ==
    PopScreen();
  End;
End;

//==============================================================================
//== Verifie si un joueur  effectu un High-Score                  ==
//== Si c'est le cas, on demande son nom et sauvegarde sur internet ==
function CheckHiScore(Player:integer;Score:integer):boolean;
Var
  i: integer;
  s: String;
  Name: String;
  HName: String;
  HSc: String;
Begin
  Result := FALSE;
  For i := 0 To HS.Count - 1 Do
  Begin
    s := HS.Strings[i];
    HName := Copy(s, 1, Pos('=', s));
    HSc := Copy(s, Pos('=', s) + 1, Length(s) - Pos('=', s));
    If Score > StrToInt(HSc) Then
    Begin
      Result := True;
      Name := GetHighScoreNickName(i,Player);
      HS.Insert(i, Name + '=' + IntToStr(Score));
      HS.Delete(HS.Count - 1);
      break;
    End;
  End;
  If Result Then
  Begin
    //== Soumet le nouveau score sur internet ==
    If INTERNET_HS_ENABLE Then SaveInternetHS();
  End;
End;


End.

