//==============================================================================
//==============================================================================
//=== MonitorHS.exe
//=== Eric Boisvert
//=== Explication sommaire du programme:
//===
//=== Un High Score arrive par internet via un serveur FTP.
//=== Le serveur FTP et ce programme sont sur la mme machine physique.
//=== Le nom du fichier HS prend la forme d'un GUID (Globally Unique IDentifiers)
//=== pour plus d'info :
//=== http://fr.wikipedia.org/wiki/Globally_Unique_Identifier
//=== Voici un exemple d'un nom de fichier GUID:
//=== {03F30B9E-803E-4837-8DA0-DCC6ED602FE1}
//=== l'utilisation des GUID comme nom de fichier garantie la non collision
//=== de nom de fichier entre 2 soumissions.
//===
//=== Ce programme fait du monitoring sur le repertoire o les fichiers
//=== HS(High Score) des internautes arrivent.
//=== Il les traite ensuite un  un et met  jour le fichier HS_NET qui contient
//=== les HS officiels pour tout les joueurs.
//=== Lorsqu'un fichier est trait, il est automatiquement effac. La disparition
//=== du fichier sur le site FTP donnera au programme "soumissionneur de HS"
//=== le signale dont il a besoin pour tlcharger  nouveau le fichier
//=== officiel des HS.
//===
//=== C'est trs basic comme approche mais trs suffissant pour maintenir
//=== une liste de HS officiel entre quelques amis...
//===
//==============================================================================
//==============================================================================

Unit main;

Interface

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ExtCtrls, ComCtrls, ShlObj, inifiles, filectrl;

{==============================================================================}
{                                }Const {                                      }
{==============================================================================}
  //== API CreateFile (Access right) declaration manquante ==
  FILE_LIST_DIRECTORY = 1;
  //== Nom du fichier qui contient les HS Officiels ==
  HI_SC_NAME = 'HS_NET';


{==============================================================================}
{                                }Type {                                       }
{==============================================================================}
  //=== Declaration manquante dans windows.pas dans D5 ===
  //=== Pour utiliser l'api: ReadDirectoryChangesW ===
  PFile_Notify_Information = ^TFileNotifyInformation;
  _FILE_NOTIFY_INFORMATION = Record
    NextEntryOffset: DWORD;
    Action: DWORD;
    FileNameLength: DWORD;
    FileName: Array[0..512] Of WCHAR;
  End;
  //  {$EXTERNALSYM _OFSTRUCT}
  TFileNotifyInformation = _FILE_NOTIFY_INFORMATION;
  FILE_NOTIFY_INFORMATION = _FILE_NOTIFY_INFORMATION;

  //== Main form ==
  TfrmMain = Class(TForm)
    Timer1: TTimer;
    txtLogger: TRichEdit;
    Panel1: TPanel;
    Label1: TLabel;
    txtDirectoryName: TEdit;
    cmdBrowse: TBitBtn;
    Panel2: TPanel;
    cmdStartMonitor: TBitBtn;
    cmdStopMonitor: TBitBtn;
    Panel3: TPanel;
    lblStatus: TLabel;
    Panel4: TPanel;
    Label2: TLabel;
    txtHS: TRichEdit;
    Procedure FormCreate(Sender: TObject);
    Procedure FormDestroy(Sender: TObject);
    Procedure cmdBrowseClick(Sender: TObject);
    Procedure cmdStartMonitorClick(Sender: TObject);
    Procedure cmdStopMonitorClick(Sender: TObject);
    Procedure Timer1Timer(Sender: TObject);
    Procedure MergeHS(FileName: String);
  private
    { Private declarations }
    //=== Directory Monitor ==
    FileNotify: Array[0..128] Of TFileNotifyInformation;
    Overlap: TOverlapped;
    hDirectoryFileHandle: HFILE;
    FileToCheck: TStringList;
    //== High Score Holder ==
    HS: TStringList;
    //== Feed-Back ==
    MonitorState: boolean;
  public
    { Public declarations }
  End;

{==============================================================================}
{                                }Var {                                        }
{==============================================================================}
  frmMain: TfrmMain;

  
{==============================================================================}
{                            }Implementation {                                 }
{==============================================================================}

{$R *.DFM}



//==============================================================================
//== Creation...
//==============================================================================
Procedure TfrmMain.FormCreate(Sender: TObject);
Var
  iniFile: TInifile;
  tempStr: String;
Begin

  //== Lecture des parametre par defaut ===
  iniFile := TInifile.Create(ChangeFileExt(application.ExeName, '.ini'));
  Try
    TempStr := iniFile.ReadString('Config', 'MonitorPath', '');
    txtDirectoryName.Text := TempStr;
  Finally
    iniFile.Free;
  End;
  //== Liste des fichiers  traiter...==
  FileToCheck := TStringList.Create;
  //== Liste des High Score ==
  HS := TStringList.Create;

  txtHS.Lines.Clear;
  txtLogger.Lines.Clear;

  lblStatus.Caption := '';
  txtLogger.Align := AlClient;
End;


//==============================================================================
//=== Destruction...
//==============================================================================
Procedure TfrmMain.FormDestroy(Sender: TObject);
Var
  iniFile: TInifile;
Begin
  //== Arrt des traitements ==
  Timer1.Enabled := FALSE;

  //== Sauvegarde de la config ==
  iniFile := TInifile.Create(ChangeFileExt(application.ExeName, '.ini'));
  Try
    iniFile.WriteString('Config', 'MonitorPath', txtDirectoryName.Text);
  Finally
    iniFile.Free;
  End;
  //== Arrt du monitoring de repertoire ==
  If MonitorState Then cmdStopMonitorClick(Sender);
  //== Effacement de la liste des fichiers  traiter ==
  FileToCheck.Clear;
  FileToCheck.Free;
  //== Efface les High Score en memoire ==
  HS.Clear;
  Hs.Free;
End;


//==============================================================================
//=== Selection d'un repertoire avec possibilit de crer des nouveaux
//=== need  ShlObj in uses
//==============================================================================
Function BrowseDialog(Const Title: String; Owner: HWND): String;
Const
  //use the new dialog layout
  BIF_NEWDIALOGSTYLE = $00000040;
Var
  lpItemID: PItemIDList;
  BrowseInfo: TBrowseInfo;
  DisplayName: Array[0..MAX_PATH] Of char;
  TempPath: Array[0..MAX_PATH] Of char;
Begin
  Result := '';
  FillChar(BrowseInfo, sizeof(TBrowseInfo), #0);
  With BrowseInfo Do Begin
    hwndOwner := Owner;
    pszDisplayName := @DisplayName;
    lpszTitle := PChar(Title);
    ulFlags := BIF_RETURNONLYFSDIRS + BIF_NEWDIALOGSTYLE;
  End;
  lpItemID := SHBrowseForFolder(BrowseInfo);
  If lpItemId <> Nil Then Begin
    SHGetPathFromIDList(lpItemID, TempPath);
    Result := TempPath;
    GlobalFreePtr(lpItemID);
  End;
End;

//==============================================================================
//=== L'usager veux selectionner un rpertoire pour faire le monitoring...
//==============================================================================
Procedure TfrmMain.cmdBrowseClick(Sender: TObject);
Var
  sTitle, sFolder: String;
Begin
  sTitle := 'Please, Tell-me where HS come from...';
  sFolder := BrowseDialog(sTitle, frmMain.Handle);
  If sFolder <> '' Then
  Begin
    txtDirectoryName.Text := sFolder;
  End;
End;


//==============================================================================
//=== Demarrage du monitoring d'un repertoire...
//==============================================================================
Procedure TfrmMain.cmdStartMonitorClick(Sender: TObject);
Var
  hNewDataEvent: Cardinal;
  i: Cardinal;
  Rep: DWORD;
  FilePathName: String;
  sr: TSearchRec;
  FileAttrs: integer;
Begin
  //=== Verifie que le repertoire existe toujours ===
  If Not DirectoryExists(txtDirectoryName.Text) Then
  Begin
    Application.MessageBox('Directory doesn''t exist!', 'Error', MB_OK);
    exit;
  End;

  //== Creation d'une liste de HS vide ==
  hs.clear;
  For i := 1 To 10 Do
  Begin
    hs.Add('Empty ' + IntToStr(i) + '=0');
  End;
  //== Si il y a deja un HS disponible dans le repertoire, on le load en mmoire
  //== et crase les vides...
  FilePathName := txtDirectoryName.Text + '\' + HI_SC_NAME;
  If FileExists(FilePathName) Then HS.LoadFromFile(FilePathName);

  //=== Nettoyage de la structure Overlap et initialisation  zro des nom
  //=== de fichier. (directory monitoring stuff)
  FillChar(overlap, sizeof(overlap), 0);
  FillChar(FileNotify[0], sizeof(FileNotify), 0);
  //=== Essaie d'ouvrire un Event avec le nom::WatchDirHSChangeEvent0-99999 ===
  //=== Ca nous en prend un qui n'existe pas! ===
  i := 0;
  hNewDataEvent := OpenEvent(EVENT_ALL_ACCESS, TRUE, PChar('WatchDirHSChangeEvent' + IntToStr(i)));
  While (hNewDataEvent <> 0) Do
  Begin
    //=== Le Event existe car on peut l'ouvrire! ===
    CloseHandle(hNewDataEvent);
    //=== Ca se produit lorsque le moniteur fonctionne sur un autre instance ===
    //=== dans le cas ou on fait du monitoring sur un autre repertoire de HS ===
    //=== Essaie d'en trouver un qui n'existe pas! ===
    inc(i);
    hNewDataEvent := OpenEvent(EVENT_ALL_ACCESS, TRUE, PChar('WatchDirDataChangeEvent' + IntToStr(i)))
  End;
  //=== Ok, Ici nous-en avons un qui n'existe pas...alors on le crer! ===
  overlap.hEvent := CreateEvent(Nil, FALSE, FALSE, PChar('WatchDirDataChangeEvent' + IntToStr(i)));
  assert(overlap.hEvent <> 0, 'Cannot Create hEvent!');

  //=== Obtient un handle sur un rpertoire...===
  hDirectoryFileHandle := CreateFile(
    PChar(txtDirectoryName.Text),
    FILE_LIST_DIRECTORY,
    FILE_SHARE_READ + FILE_SHARE_WRITE + FILE_SHARE_DELETE,
    Nil,
    OPEN_EXISTING,
    FILE_FLAG_BACKUP_SEMANTICS + FILE_FLAG_OVERLAPPED + FILE_FLAG_NO_BUFFERING,
    0);
  assert(hDirectoryFileHandle <> 0, 'Cannot Open Directory for monitoring!');

  //=== Demande  windows de nous avertire si il y a un changement dans ce rpertoire ===
  MonitorState := ReadDirectoryChangesW(hDirectoryFileHandle
    , @FileNotify[0]
    , SizeOf(FileNotify)
    , FALSE // Monitor Subdirectory?
    , FILE_NOTIFY_CHANGE_LAST_WRITE
    , @Rep
    , @Overlap
    , Nil); //@FileIoCompletionRoutine );
  Assert(MonitorState = TRUE, 'Cannot Monitor directory! ReadDirectoryChangeW() Call Fail!');

  //== donne un feed-back visuelle
  cmdStartMonitor.Enabled := FALSE;
  cmdStopMonitor.Enabled := TRUE;
  txtDirectoryName.ReadOnly := TRUE;
  txtLogger.Lines.Add(FormatDateTime('yyyy-mm-dd hh:nn:ss', Now()) + '  HS Monitor is ready for :' + txtDirectoryName.Text);

  //=== Verifie si il y a deja des fichiers non trait ===
  FileAttrs := faAnyFile Xor faDirectory;
  If FindFirst(txtDirectoryName.Text + '\{*.*', FileAttrs, sr) = 0 Then
  Begin
    Try
      If sr.Name <> HI_SC_NAME Then
      Begin
        txtLogger.Lines.Add('Old files processing...');
        FileToCheck.Add(sr.name);
      End;
      While FindNext(sr) = 0 Do
      Begin
        If sr.Name <> HI_SC_NAME Then
        Begin
          FileToCheck.Add(sr.name);
        End;
      End;
    Finally
      FindClose(sr);
    End;
  End;
End;


//==============================================================================
//=== Arrt du HS monitoring...
//==============================================================================
Procedure TfrmMain.cmdStopMonitorClick(Sender: TObject);
Var
  FilePathName: String;
Begin
  //== Feed-Back ===
  MonitorState := FALSE;
  txtLogger.Lines.Add(FormatDateTime('yyyy-mm-dd hh:nn:ss', Now()) + '  HS Monitor is Off:' + txtDirectoryName.Text);

  //=== Fermeture du handle du repertoire ===
  If hDirectoryFileHandle <> 0 Then
  Begin
    CloseHandle(hDirectoryFileHandle);
    hDirectoryFileHandle := 0;
  End;
  //=== Fermeture du Event Handle ==
  If overlap.hEvent <> 0 Then
  Begin
    CloseHandle(overlap.hEvent);
    overlap.hEvent := 0;
  End;

  //=== Sauvegarde des HS dans le repertoire ===
  FilePathName := txtDirectoryName.Text + '\' + HI_SC_NAME;
  HS.SaveToFile(FilePathName);
  HS.Clear;

  //=== Redonne le controle  l'usager pour des changements et/ou un redemarrage
  txtDirectoryName.ReadOnly := FALSE;
  cmdStopMonitor.Enabled := FALSE;
  cmdStartMonitor.Enabled := TRUE;
End;


//==============================================================================
//=== Timer  toute les 500ms
//=== Traitement et Feed-Back sur l'tat du serveur...
//==============================================================================
Procedure TfrmMain.Timer1Timer(Sender: TObject);
Var
  Rep: DWORD;
  nbrbytes: cardinal;
  i: cardinal;
  TempString: String;
Begin
  If MonitorState Then
  Begin
    lblStatus.Caption := 'Monitor is ON!';
    txtHS.Text := hs.Text;
  End
  Else
  Begin
    lblStatus.Caption := 'Monitor is OFF!';
    Exit;
  End;

  //=== Traite les fichiers detects le coup d'avant... ===
  If FileToCheck.Count <> 0 Then
  Begin
    For i := FileToCheck.Count - 1 Downto 0 Do
    Begin
      TempString := FileToCheck.Strings[i];
      If (pos('{', TempString) <> 0) And (pos('}', TempString) <> 0) Then
      Begin
        //=== Il s'agit d'une soumission d'un HS par un Internaute... ===
        txtLogger.Lines.Add(FormatDateTime('yyyy-mm-dd hh:nn:ss', Now()) + '  Processing file:' + TempString);
        MergeHS(TempString);
        //=== Si le fichier n'existe plus, le traitement c'est bien pass ==
        //=== On peut alors enlever le fichier de la liste  traiter ==
        If Not FileExists(txtDirectoryName.Text + '\' + TempString) Then
          FileToCheck.Delete(i);
      End;
    End;
  End;

  //=== Verifie si windows nous indiques qu'il y a du changement dans ==
  //=== le repertoire des HS ==
  If WaitForSingleObject(overlap.hEvent, 0) = WAIT_OBJECT_0 Then
  Begin
    //=== Il y a des changements dans le repertoire! ===
    GetOverlappedResult(hDirectoryFileHandle
      , overlap
      , nbrbytes
      , FALSE);
    ResetEvent(overlap.hEvent);
    If nbrbytes <> 0 Then
    Begin
      //=== Lecture des noms de fichiers (Il y en a peut-tre plus d'un!) ===
      For i := 0 To trunc(nbrBytes / sizeOf(TFileNotifyInformation)) Do
      Begin
        //=== Obtien le nom du fichier qui vient d'arriver ===
        //=== dans le repertoire de HS ==
        TempString := trim(WideCharToString(FileNotify[i].FileName));
        If FileExists(txtDirectoryName.Text + '\' + TempString) Then
        Begin
          txtLogger.Lines.Add(FormatDateTime('yyyy-mm-dd hh:nn:ss', Now()) + '  New File Detected:' + TempString);
          FileToCheck.Add(TempString);
        End;
      End;
    End;
    //=== Demande  Windows de nous avertire ENCORE si il y a des changements ===
    //=== dans ce repertoire ===
    FillChar(FileNotify[0], sizeof(FileNotify), 0);
    MonitorState := ReadDirectoryChangesW(hDirectoryFileHandle
      , @FileNotify[0]
      , SizeOf(FileNotify)
      , LongBool(0), FILE_NOTIFY_CHANGE_LAST_WRITE
      , @Rep
      , @Overlap
      , Nil);
  End;
End;

//==============================================================================
//=== MergeHS tente de lire et de fusionner le fichier HS de l'internaute
//=== avec celui en mmoire. si tout va bien, le HS de l'internaute est effac.
//=== et un nouveau fichier HS Officiel est enregistrer dans le rpertoire.
//=== Si le fichier HS est incomplet, on ne fait rien et il sera traiter au
//=== 500ms prochain...
//==============================================================================
Procedure TfrmMain.MergeHS(FileName: String);
Var
  sl: TStringList;
  i, j: integer;
  s: String;
  Name: String;
  Sc: String;
  HName: String;
  HSc: String;
Begin
  sl := TStringList.Create;
  Try
    If FileExists(txtDirectoryName.Text + '\' + FileName) Then
    Begin
      sl.LoadFromFile(txtDirectoryName.Text + '\' + FileName);
      If sl.count = 10 Then
      Begin
        For i := 0 To sl.Count - 1 Do
        Begin
          s := sl.Strings[i];
          Name := Copy(s, 1, Pos('=', s));
          sc := Copy(s, Pos('=', s) + 1, Length(s) - Pos('=', s));
          For j := 0 To HS.Count - 1 Do
          Begin
            s := HS.Strings[j];
            HName := Copy(s, 1, Pos('=', s));
            HSc := Copy(s, Pos('=', s) + 1, Length(s) - Pos('=', s));
            If (Name = HName) And (StrToInt(sc) = StrToInt(HSc)) Then break;
            If StrToInt(sc) > StrToInt(HSc) Then
            Begin
              HS.Insert(j, Name + sc);
              HS.Delete(HS.Count - 1);
              break;
            End;
          End;
        End;
        //=== Save New score ===
        hs.SaveToFile(txtDirectoryName.Text + '\' + HI_SC_NAME);
        //=== Efface le fichier de l'internaute ==
        DeleteFile(txtDirectoryName.Text + '\' + FileName);
      End;
    End;
  Finally
    sl.Clear;
    sl.free;
  End;
End;

End.

